#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
import os
import shutil
import copy
import sys
import inspect

import numpy as np

from .inlists import defaults, evaluate_inlist, variable_to_string
from ..functions import get_mesa_version, compare_version

non_mesa_key_start = '!PY_KEY_'
kap_user_params = ['user_num_kap_Xs', 'user_kap_Xs', 'user_num_kap_Zs', 'user_kap_Zs',
                   'user_num_kap_Xs_for_this_Z',
                   'user_num_kap_CO_Xs', 'user_kap_CO_Xs', 'user_num_kap_CO_Zs', 'user_kap_CO_Zs',
                   'user_num_kap_CO_Xs_for_this_Z',
                   'user_num_kap_lowT_Xs', 'user_kap_lowT_Xs', 'user_num_kap_lowT_Zs', 'user_kap_lowT_Zs',
                   'user_num_kap_lowT_Xs_for_this_Z']
class MesaGrid:
    def __init__(self, mesa_dir='', inlist_filename='inlist', inlists_index=5, starjob_filename='inlist_project',
                 controls_filename='inlist_project', eos_filename='inlist_project', kap_filename='inlist_project',
                 pgstar_filename='inlist_project', add_base_workdir=False):
        """
        `MesaGrid` class which contains all inlist settings for a grid. When the `create_grid` method is called,
        a copy of the script which called it is copied into the grid directory.

        Args:
            mesa_dir (str): ``$MESA_DIR`` root directory to be used with this grid.
            inlist_filename (str, optional): Defaults to 'inlist'.
            inlists_index (int, optional): Which read_extra_*_inlist(inlists_index) to use. Defaults to 5.
            starjob_filename (str, optional): Cannot be the same as `inlist_filename`. Defaults to 'inlist_project'.
            controls_filename (str, optional): Cannot be the same as `inlist_filename`. Defaults to 'inlist_project'.
            eos_filename (str, optional): Cannot be the same as `inlist_filename`. Defaults to 'inlist_project'.
            kap_filename (str, optional): Cannot be the same as `inlist_filename`. Defaults to 'inlist_project'.
            pgstar_filename (str, optional): Cannot be the same as `inlist_filename`. Defaults to 'inlist_project'.
            add_base_workdir (bool, optional): Add minimum required files from $MESA_DIR/star/work to each run directory (make/, src/, mk, re, rn, clean).

        Attributes:
            star_job (dict): Options for ``star_job``.
            controls (dict): Options for ``controls``.
            pgstar (dict): Options for ``pgstar``.
            kap (dict): Options for ``kap``. Only exists if `mesa_dir` version is later than or equal to 15140.
            eos (dict): Options for ``eos``. Only exists if `mesa_dir` version is later than or equal to 15140.

        Examples:
            Creating a grid of MESA runs with initial masses of 1 and 2 Msol.

            >>> import os
            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.controls['initial_mass'] = [1, 2]
            >>> grid.create_grid('path/to/grid')
            >>> os.listdir('path/to/grid')
            ['0000', '0001]
        TODO:
            * Add astero namelist and include in tests.
        """

        if inlist_filename in [starjob_filename, controls_filename, eos_filename, kap_filename, pgstar_filename]:
            raise ValueError(
                '`inlist_filename` cannot be the same as `starjob_filename`, `controls_filename`, `eos_filename`, `kap_filename`, or `pgstar_filename`.')

        if mesa_dir == '':
            self.mesa_dir = os.environ['MESA_DIR']
            self.version = get_mesa_version(self.mesa_dir)
        else:
            self.version = get_mesa_version(mesa_dir)
            self.mesa_dir = mesa_dir

        if compare_version(self.version, '15140', '>='):
            self.namelists = ('star_job', 'eos', 'kap', 'controls', 'pgstar')
            self.separate_eoskap = True
        else:
            self.namelists = ('star_job', 'controls', 'pgstar')
            self.separate_eoskap = False

        if compare_version(self.version, 'r23.05.1', '>='):
            self.extra_inlist_as_list = True
        else:
            self.extra_inlist_as_list = False

        # Setup namelist objects
        self.star_job = {f'{non_mesa_key_start}filename': starjob_filename,
                         f'{non_mesa_key_start}type': 'star_job',
                         f'{non_mesa_key_start}group_unpack': []}
        self.controls = {f'{non_mesa_key_start}filename': controls_filename,
                         f'{non_mesa_key_start}type': 'controls',
                         f'{non_mesa_key_start}group_unpack': []}
        # self.astero = {f'{non_mesa_key_start}filename': astero_filename,
        #                  f'{non_mesa_key_start}type': 'astero',
        #                  f'{non_mesa_key_start}group_unpack': []}
        if self.separate_eoskap:
            self.eos = {f'{non_mesa_key_start}filename': eos_filename,
                        f'{non_mesa_key_start}type': 'eos',
                        f'{non_mesa_key_start}group_unpack': []}
            self.kap = {f'{non_mesa_key_start}filename': kap_filename,
                        f'{non_mesa_key_start}type': 'kap',
                        f'{non_mesa_key_start}group_unpack': []}
        self.pgstar = {f'{non_mesa_key_start}filename': pgstar_filename,
                       f'{non_mesa_key_start}type': 'pgstar', }
        self.inlist = {f'{non_mesa_key_start}filename': inlist_filename,
                       f'{non_mesa_key_start}type': 'master',
                       f'{non_mesa_key_start}group_unpack': []}

        self.inlists_index = inlists_index
        for namelist in self.namelists:
            namelist_filename = self.__dict__[namelist][f'{non_mesa_key_start}filename']
            i = self.inlists_index
            if self.extra_inlist_as_list:
                read_extra_inlist_key = f'read_extra_{namelist}_inlist({i})'
                read_extra_inlist_name_key = f'extra_{namelist}_inlist_name({i})'
            else:
                read_extra_inlist_key = f'read_extra_{namelist}_inlist{i}'
                read_extra_inlist_name_key = f'extra_{namelist}_inlist{i}_name'

            # Create entry and default to True.
            self.inlist[namelist] = {read_extra_inlist_key: True,
                                     read_extra_inlist_name_key: namelist_filename}

        self.extra_files = []
        self.extra_dirs = []
        self.inlist_finalize_function = None
        self.griddir_finalize_function = None
        self.name_funcion = None
        self.unpacked = False

        self.inlist_option_files_to_validate = [('star_job', 'history_columns_file'),
                                                ('star_job', 'profile_columns_file')]

        if add_base_workdir:
            for fname in ['mk', 'clean', 'rn', 're']:
                self.extra_files.append(os.path.join(f'{self.mesa_dir}/star/work', fname))
            for dname in ['make', 'src']:
                self.extra_dirs.append(os.path.join(f'{self.mesa_dir}/star/work', dname))

        no_expand = {namelist: [] for namelist in self.namelists}
        no_expand['kap'] = ['user_kap_Xs', 'user_kap_Zs', 'user_num_kap_Xs_for_this_Z', 'user_kap_CO_Xs',
                             'user_kap_CO_Zs', 'user_num_kap_CO_Xs_for_this_Z', 'user_kap_lowT_Xs', 'user_kap_lowT_Zs',
                             'user_num_kap_lowT_Xs_for_this_Z']
        no_expand['kap'] = [_.lower() for _ in no_expand['kap']]
        self.no_expand = no_expand
        self.expand_non_mesa_keys = {namelist:[] for namelist in self.namelists}

    def __repr__(self):
        s = f'MESA Grid for version {self.version}.'
        if self.unpacked:
            s += f' {len(self.unpacked)} runs.'
        return s

    def add_file(self, path):
        """
        Add a file which will be copied into each run directory. Will raise a FileNotFoundError if the file at
        `path` does not exist.

        Args:
            path (str): Path to file to be added to every run directory.

        Examples:
            Copy the file at ``path/to/file`` to each run directory when the grid is created.

            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.add_file('path/to/file')
        """
        if os.path.isfile(path):
            self.extra_files.append(path)
        else:
            raise FileNotFoundError(f'{path} is not a file.')

    def add_dir(self, path):
        """
        Add a directory which will be copied into each run directory.

        Args:
            path (str): Path to directory to be added to every run directory.

        Examples:
            Copy the directory at ``path/to/directory`` to each run directory when the grid is created.

            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.add_dir('path/to/directory')
        """
        if os.path.isdir(path):
            self.extra_dirs.append(path)
        else:
            raise NotADirectoryError(f'{path} is not a directory.')

    def add_inlist_option_file_check(self, namelist, option):
        """
        Add a check that the file specified in the `namelist[option]` must exist in each run directory,
        e.g. controls['history_columns_filename'], which is included by default as well as ``profile_columns_filename``.

        Args:
            namelist (str):
            option (str):

        Examples:

            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.add_inlist_option_file_check('controls', 'history_columns_file')

        """
        if namelist not in self.namelists:
            raise ValueError(f'`namelist` {namelist} must be one of {", ".join(self.namelists)}.')
        self.inlist_option_files_to_validate.append((namelist, option))


    def add_no_expand_key(self, namelist, key):
        """
        Add a key which will not be expanded. For example &kap's ``user_kap_Xs``
        Args:
            namelist (str): Which namelist the key is part of. E.g. 'kap'.
            key (str): Name of parameter. E.g. 'user_kap_Xs'.

        Examples:

            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.add_no_expand_key('kap', f'user_kap_Xs')

        """

        self.no_expand[namelist].append(key)

    def add_non_mesa_key(self, namelist, key):
        """
        Add a key which is not in MESA's keys and expand it. For example including a parameter which is used in
        ``inlist_finalize_function``.
        Args:
            namelist (str): Which namelist the key is part of. E.g. controls.
            key (str): Name of parameter. Must start with ``'!PY_KEY_'``.
        """

        if not key.startswith(non_mesa_key_start):
            raise ValueError(f'Key {key} must start with "{non_mesa_key_start}"')

        self.expand_non_mesa_keys[namelist].append(key)

    def set_inlist_finalize_function(self, function):
        """
        Set the function which is applied to all unpacked inlists.

        The function must accept a single `inlist` and return a single `inlist`.

        Args:
            function (function):

        Examples:
            Change an option in `star_job` depending on ``initial_mass`` in ``controls``. This could also be accomplished using the
            ``f'{non_mesa_key_start}group_unpack'`` key.
            
            >>> import numpy as np
            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.star_job['change_initial_net'] = True
            >>> grid.star_job['new_net_name'] = 'pp_extras.net'
            >>> grid.controls['initial_mass'] = np.linspace(1, 2, 6)
            >>> def finalize_function(unpacked_inlist):
            ...     if unpacked_inlist['controls']['initial_mass'] > 1.3:
            ...         unpacked_inlist['star_job']['new_net_name'] = 'pp_cno_extras.net'
            ...     return unpacked_inlist
            >>> grid.set_inlist_finalize_function(finalize_function)
            >>> grid.unpack_inlists()
            >>> grid.unpacked[0]['star_job']['new_net_name']
            'pp_extras.net'
            >>> grid.unpacked[-1]['star_job']['new_net_name']
            'pp_cno_extras.net'

        """
        self.inlist_finalize_function = function

    def set_griddir_finalize_function(self, function):
        """
        Set the function which is called in each grid directory. The arguments to the function are a MesaGrid object and
        the run index.

        Args:
            function (function):

        Examples:

            >>> import os
            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.controls['initial_mass'] = [1, 2]
            >>> def griddir_finalize_function(grid, i):
            ...     os.system("pwd")
            >>> grid.set_griddir_finalize_function(griddir_finalize_function)
            >>> grid.create_grid('path/to/grid')
            /home/walter/Github/wsssss/path/to/grid/0000
            /home/walter/Github/wsssss/path/to/grid/0001

        """
        self.griddir_finalize_function = function

    def set_name_function(self, function):
        """
        Set the function which is used to generate the name of each run directory.
        It recieves a fully unpacked inlist (dict of dicts) or grid index and unpacked inlist and returns a string.
        It should generate a unique name for each inlist.

        Args:
            function (function):

        Examples:

            >>> import os
            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.controls['initial_mass'] = [1, 2]
            >>> def name_function(unpacked_inlist):
            ...     return f'm{unpacked_inlist["controls"]["initial_mass"]:.3f}'
            >>> grid.set_name_function(name_function)
            >>> grid.create_grid('path/to/grid')
            >>> os.listdir('path/to/grid')
            ['m1.000', 'm2.000']
        """
        self.name_funcion = function

    def create_grid(self, grid_path, rm_dir=True):
        """
        Create the MESA grid in `grid_path`. The inlist options are validated and extra files and directories copied to
        each run directory.
        Args:
            grid_path: Path to grid directory.
            rmdir (bool, optional): Remove grid directory if it exists. Defaults to `True`.

        Examples:

            >>> import os
            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.controls['initial_mass'] = [1, 2]
            >>> grid.create_grid('path/to/grid')
            >>> os.listdir('path/to/grid')
            ['0000', '0001']
        """
        self.validate_inlists()

        self.unpack_inlists()

        self.check_copy()

        self._setup_directories(grid_path, rm_dir)

        self._write_inlists(grid_path)

        self._copy_extra_files_and_dirs(grid_path)

        curdir = os.path.abspath('.')
        abs_grid_path = os.path.abspath(grid_path)
        for i, dirname in enumerate(self.dirnames):
            os.chdir(os.path.join(abs_grid_path, dirname))
            if self.griddir_finalize_function is not None:
                self.griddir_finalize_function(self, i)
        os.chdir(curdir)

        self._validate_files(grid_path)

        # Copy file which called create_grid into grid_path
        if len(inspect.stack()) > 1:
            calling_file = inspect.stack()[1].filename
            if os.path.isfile(calling_file) and not calling_file.endswith('IPython/core/interactiveshell.py'):
                shutil.copy2(calling_file, grid_path)
            else:
                print(f"Cannot copy calling file {calling_file} to grid directory {grid_path}.")
        else:
            pass


    def validate_inlists(self, mesa_dir=None):
        """
        Check if all options in the inlists are valid MESA keys.

        Args:
            mesa_dir (str, optional): MESA root directory to check against. Defaults to None, which will use ``$MESA_DIR``.

        Examples:

            >>> from wsssss.inlists import create_grid as cg
            >>> grid = cg.MesaGrid()
            >>> grid.controls['this_option_does_not_exist'] = 1
            KeyError: 'Option(s) not in available controls keys: this_option_does_not_exist.'
        """
        # Check if all extra namelists filenames are unique and different from the main inlist name.
        for namelist in self.namelists:
            read_extra_names = []
            for i in range(1, 6):
                if self.extra_inlist_as_list:
                    read_extra_inlist_key = f'read_extra_{namelist}_inlist({i})'
                    read_extra_inlist_name_key = f'extra_{namelist}_inlist_name({i})'
                else:
                    read_extra_inlist_key = f'read_extra_{namelist}_inlist{i}'
                    read_extra_inlist_name_key = f'extra_{namelist}_inlist{i}_name'

                if read_extra_inlist_key not in self.inlist[namelist].keys():
                    continue

                if self.inlist[namelist][read_extra_inlist_key]:
                    read_extra_names.append(self.inlist[namelist][read_extra_inlist_name_key])
                    if self.inlist[f'{non_mesa_key_start}filename'] == self.inlist[namelist][
                        read_extra_inlist_name_key]:
                        raise ValueError(
                            f'Main inlist filename {self.inlist[f"{non_mesa_key_start}filename"]} cannot be the same as an extra inlist filename {self.inlist[namelist][read_extra_inlist_name_key]}.')

            if len(read_extra_names) != len(np.unique(read_extra_names)):
                raise ValueError(f'Extra {namelist} inlist names must be unique.')

        # Get available MESA options from mesa_dir
        if mesa_dir is None:
            mesa_dir = self.mesa_dir
        available_options = {}
        for namelist in self.namelists:
            available_options[namelist] = self._get_available_options(f'{mesa_dir}/{defaults[namelist]}')  # TODO: replace with inlists.get_mesa_defaults

        # Find which options are not in MESA options.
        failed_options = {}
        failed = False
        for namelist in self.namelists:
            failed_options[namelist] = []
            for key in self.__dict__[namelist].keys():
                if not key.startswith(non_mesa_key_start):
                    if key.endswith(')'):
                        compare_key = key.split('(')[0]  # Cut off array index part
                    else:
                        compare_key = key
                    if compare_key.lower() not in available_options[namelist]:
                        if namelist == 'kap':  # These options are commented out in kap.defaults
                            if compare_key.lower() in [_.lower() for _ in kap_user_params]:
                                continue
                        failed_options[namelist].append(key)
                        failed = True

        if failed:
            s = []
            for namelist in self.namelists:
                if len(failed_options[namelist]) > 0:
                    s.append(f'Option(s) not in available {namelist} keys: {", ".join(failed_options[namelist])}.')
            s = ' '.join(s)
            raise KeyError(s)

    def _validate_files(self, grid_path):
        """
        Check if all specified files are accounted for.
        Args:
            grid_path (str): Grid directory.

        """

        file_not_found = []

        for i, dirname in enumerate(self.dirnames):
            run_dir = os.path.join(grid_path, dirname)
            inlist = evaluate_inlist(os.path.join(run_dir, self.inlist[f'{non_mesa_key_start}filename']))
            for (kind, key) in self.inlist_option_files_to_validate:
                if kind in inlist.keys():
                    if key in inlist[kind].keys():
                        fname = inlist[kind][key]
                        if not os.path.isfile(os.path.join(run_dir, fname)):
                            file_not_found.append(fname)

        s = ''
        if file_not_found:
            s += f'Missing {len(file_not_found)} files: ' + ' '.join(file_not_found)
            s += '\n'

        if s:
            raise FileNotFoundError(s)

    def _get_available_options(self, path):
        """
        TODO: Replace with inlists.get_mesa_defaults
        Args:
            path:

        Returns:

        """
        with open(path, 'r') as handle:
            lines = handle.readlines()
        lines = [_.strip() for _ in lines]
        lines = [_ for _ in lines if not _.startswith('!') and _]
        lines = [_.split('=')[0].strip() for _ in lines]
        lines = [_.lower() for _ in lines]
        lines = [_.split('(')[0] for _ in lines]
        return lines

    def check_copy(self):
        """
        Check if all files and directories which are to be copied exist.

        """

        file_not_found = []
        for fpath in self.extra_files:
            if not os.path.isfile(fpath):
                file_not_found.append(fpath)

        dir_not_found = []
        for dpath in self.extra_dirs:
            if not os.path.isdir(dpath):
                dir_not_found.append(dpath)

        # Check for name collisions

        s = ''
        if file_not_found:
            s += f'Missing {len(file_not_found)} files: ' + ' '.join(file_not_found)
            s += '\n'

        if dir_not_found:
            s += f'Missing {len(dir_not_found)} directories: ' + ' '.join(dir_not_found)

        if s:
            raise FileNotFoundError(s)

    def unpack_inlists(self):
        """
        Unpack all inlists. Will place them in the `unpacked` attribute as a tuple.

        """
        generators = []
        for namelist in ['inlist', *self.namelists]:
            generators.append(self._make_inlist_generator(self.__dict__[namelist]))

        all_unpacked = []
        for i, unpacked in enumerate(itertools.product(*generators)):
            unpacked = dict(zip(['inlist', *self.namelists], list(unpacked)))
            if self.inlist_finalize_function is not None:
                finalized_unpacked = self.inlist_finalize_function(copy.deepcopy(unpacked))
            else:
                finalized_unpacked = unpacked
            all_unpacked.append(finalized_unpacked)
        self.unpacked = tuple(all_unpacked)

        num_unpacked = len(self.unpacked)
        self.dirnames = []
        if self.name_funcion is None:
            num_digits = max(4, np.ceil(np.log10(num_unpacked)).astype(int))
            for i in range(num_unpacked):
                dirname = f'{i:0{num_digits}}'
                self.dirnames.append(dirname)
        else:
            for i in range(num_unpacked):
                if self.name_funcion.__code__.co_argcount == 1:
                    dirname = self.name_funcion(self.unpacked[i])
                else:
                    dirname = self.name_funcion(i, self.unpacked[i])
                if dirname in self.dirnames:
                    i_already_exists = self.dirnames.index(dirname)
                    raise ValueError(f'`dirname` {dirname} for inlist {i} has already been generated for inlist {i_already_exists}.')
                self.dirnames.append(dirname)
        self.dirnames = tuple(self.dirnames)

    def _make_inlist_generator(self, inlist_dict):
        """
        Create a generator which yields all unique inlists with all combinations of any lists, tuples, or numpy arrays of length greater than 1.
        Can handle groups of variables which must change together by appending a list of `dict`s in the `group_unpack` key. If multiple lists of
        `dict`s are appended, they are treated as separate variables.
        Args:
            inlist_dict (dict):

        Yields:
            dict: Unpacked inlist.

        """
        inlist_dict = copy.deepcopy(inlist_dict)
        if inlist_dict[f'{non_mesa_key_start}type'] == 'master':
            yield inlist_dict
            return
        if inlist_dict[f'{non_mesa_key_start}type'] == 'pgstar':
            yield inlist_dict
            return

        contains_list = list()  # Keys which need to be unpacked
        lengths = list()  # Length of lists which need to be unpacked
        items = list()  # Lists which need to be unpacked
        for key, item in inlist_dict.items():
            if key.lower() in self.no_expand[inlist_dict[f'{non_mesa_key_start}type']]:
                continue

            if key.startswith(non_mesa_key_start):
                if key == f'{non_mesa_key_start}group_unpack':
                    groups = inlist_dict[f'{non_mesa_key_start}group_unpack']
                    for i, group in enumerate(groups):
                        contains_list.append(f'{non_mesa_key_start}__group__{i}')
                        lengths.append(len(group))
                        items.append(group)
                        continue
                elif key in self.expand_non_mesa_keys[inlist_dict[f'{non_mesa_key_start}type']]:  # Continue as if it is a normal key and expand it
                    pass
                else:
                    continue

            if type(item) in [list, tuple, np.ndarray]:
                if len(item) > 1:
                    contains_list.append(key)
                    lengths.append(len(item))
                    items.append(inlist_dict[key])
                elif len(item) == 1:  # Remove item from list
                    inlist_dict[key] = item[0]

        if len(lengths) == 0:
            out_inlist = inlist_dict.copy()
            for namelist in self.namelists:
                if f'{non_mesa_key_start}group_unpack' in out_inlist.keys():
                    out_inlist.pop(f'{non_mesa_key_start}group_unpack')
            yield out_inlist
            return

        tot_permutations = 1
        for n in lengths:
            tot_permutations *= n

        base_inlist = copy.deepcopy(inlist_dict)
        if f'{non_mesa_key_start}group_unpack' in base_inlist.keys():
            base_inlist.pop(f'{non_mesa_key_start}group_unpack')

        permutations = itertools.product(*items)
        for i, permutation in enumerate(permutations):
            new_inlist = base_inlist.copy()
            for j, value in enumerate(permutation):
                if contains_list[j].startswith(f'{non_mesa_key_start}__group__'):
                    new_inlist.update(value)
                else:
                    new_inlist[contains_list[j]] = value
            yield new_inlist

    def _setup_directories(self, grid_path, rm_dir):
        """
        Create the directory structure for a grid at ``grid_path``.

        Args:
            grid_path (str): Path which will contain the grid.
            rm_dir (bool): If it exists, first remove the grid directory.

        """

        if os.path.isfile(grid_path):
            raise FileExistsError(f'Expected `grid_path` {grid_path} is a file.')

        if os.path.exists(grid_path) and os.path.isdir(grid_path) and rm_dir:  # Remove existing grid.
            shutil.rmtree(grid_path)

        os.makedirs(grid_path, exist_ok=True)
        for dirname in self.dirnames:
            os.makedirs(os.path.join(grid_path, dirname), exist_ok=True)


    def _write_inlists(self, grid_path):
        """
        Write the unpacked inlists to ``grid_path`` in their respective directories.

        Args:
            grid_path (str): Path which will contain the grid.

        """
        for i, dirname in enumerate(self.dirnames):
            dirpath = os.path.join(grid_path, dirname)
            unpacked = self.unpacked[i]

            # Must remove pre-existing inlist files first
            for namelist in unpacked.keys():
                file_path = os.path.join(dirpath, unpacked[namelist][f'{non_mesa_key_start}filename'])
                if os.path.exists(file_path):
                    os.remove(file_path)

            for namelist in unpacked.keys():
                file_path = os.path.join(dirpath, unpacked[namelist][f'{non_mesa_key_start}filename'])
                if not os.path.exists(file_path):
                    prepend = '!Created using wsssss.inlists.create_grid module.\n'
                else:
                    prepend = ''
                with open(file_path, 'a') as handle:
                    handle.write(prepend + self._generate_inlist_string(unpacked[namelist]))

    def _generate_inlist_string(self, inlist_dict):
        """
        Generates an inlist string from `inlist_dict`. Which is an unpacked controls etc dict.

        Args:
            inlist_dict (dict): Dictionary containing a set of inlist options.

        Returns:
            inlist_str (str): A string representation of ``inlist_dict`` readable by MESA.
        """
        inlist_string = ''
        inlist_type = inlist_dict[f'{non_mesa_key_start}type']


        if inlist_type == 'master':
            for key in inlist_dict.keys():
                if key.startswith(non_mesa_key_start):
                    continue
                elif key.startswith('#'):
                    continue
                key_dat = inlist_dict[key]
                sub_str = f'\n&{key}\n'

                for sub_key, sub_value in key_dat.items():
                    if sub_key.startswith(non_mesa_key_start):
                        continue
                    parsed_sub_value = variable_to_string(sub_value)
                    sub_str += f'    {sub_key} = {parsed_sub_value}\n'

                sub_str += rf'/ ! end of {key} namelist'
                sub_str += '\n'

                inlist_string += sub_str
        else:
            sub_str = f'\n&{inlist_type}\n'
            for key, value in inlist_dict.items():
                if key.startswith(non_mesa_key_start):
                    if key == 'note':
                        sub_str = f'! {value}\n' + sub_str
                        continue
                    else:
                        continue
                elif key.startswith('#'):
                    continue
                if key.lower() in self.no_expand[inlist_type]:
                    parsed_value = ', '.join([variable_to_string(val) for val in value])
                else:
                    parsed_value = variable_to_string(value)
                sub_str += f'    {key} = {parsed_value}\n'

            sub_str += f'/ ! end of {inlist_type} namelist\n'

            inlist_string += sub_str

        return inlist_string

    def _copy_extra_files_and_dirs(self, grid_path):
        """
        Copy extra files and directories specified in ``MesaGrid.extra_files`` and ``MesaGrid.extra_dirs`` to each
        run directory.

        Args:
            grid_path:
        """
        for dirname in self.dirnames:
            dirpath = os.path.join(grid_path, dirname)
            for fpath in self.extra_files:
                fname = os.path.basename(fpath)
                shutil.copy2(os.path.abspath(fpath), os.path.join(dirpath, fname))
            for dpath in self.extra_dirs:
                dname = os.path.basename(dpath)
                shutil.copytree(os.path.abspath(dpath), os.path.join(dirpath, dname), dirs_exist_ok=True)

            # Also copy files specified in inlist_option_files_to_validate if they are missing in the run directory.
            for namelist, option in self.inlist_option_files_to_validate:
                if not option in self.__dict__[namelist]:
                    continue
                fpath = self.__dict__[namelist][option]
                if fpath == '':
                    pass
                else:
                    fname = os.path.basename(fpath)
                    if os.path.isfile(os.path.join(dirpath, fname)):
                        pass
                    else:
                        shutil.copy2(os.path.abspath(fpath), os.path.join(dirpath, fname))

            # Copy extra inlist files if specified.
            for namelist in self.namelists:
                for i in range(5):
                    i += 1
                    if i == self.inlists_index:  # Skip files generated by this class
                        continue
                    if self.extra_inlist_as_list:
                        read_extra_inlist_key = f'read_extra_{namelist}_inlist({i})'
                        read_extra_inlist_name_key = f'extra_{namelist}_inlist_name({i})'
                    else:
                        read_extra_inlist_key = f'read_extra_{namelist}_inlist{i}'
                        read_extra_inlist_name_key = f'extra_{namelist}_inlist{i}_name'

                    if read_extra_inlist_key in self.inlist[namelist].keys():
                        if self.inlist[namelist][read_extra_inlist_key]:
                            fname = os.path.basename(self.inlist[namelist][read_extra_inlist_name_key])
                            if fname not in [os.path.basename(p) for p in self.extra_files]:
                                shutil.copy2(os.path.abspath(self.inlist[namelist][read_extra_inlist_name_key]),
                                         os.path.join(dirpath, fname))


    def summary(self):
        """
        Print a summary of which variables change.
        """
        raise NotImplementedError()
        if not self.unpacked:
            self.unpack_inlists()

        for kind in self.namelists:
            pass



import unittest

from wsssss.constants import pre15140
from wsssss.constants import post15140

class TestConstants(unittest.TestCase):
    def test_scalar(self):
        for c in [pre15140, post15140]:
            keys = list(c.__dict__.keys())
            i_start = keys.index('version')
            for key in keys[i_start:]:
                if key == 'version':
                    self.assertTrue(isinstance(c.version, str))
                else:
                    self.assertTrue(isinstance(c.__dict__[key], (float, int)), f'{key} is not int or float.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import ast
import glob
import os
import sys
from argparse import ArgumentParser
from collections import Counter
from collections import defaultdict
from datetime import datetime

import numpy as np
import warnings

from wsssss import load_data as ld
from wsssss import functions as uf

warnings.filterwarnings("error")
def get_parser():
    """"""
    _parser = ArgumentParser(description="""
    This program will check whether a grid has run successfully.
    """)
    _parser.add_argument('--verbose', '-v', action='store_const', const=True, default=False, )
    _parser.add_argument('grid_dir', type=str, default='.', nargs='?',
                         help='The directory which contains each run for MESA to run in a separate sub-directory.')
    _parser.add_argument('--subdirs', '-d', nargs='*', type=str, default='',
                         help='The names of the sub directories if they are not named 0000 0001 etc.')
    _parser.add_argument('--history-file', '-f', type=str, default='history.data',
                         help='Name of the MESA history file in the LOGS directory. {} will be expanded to the '
                              'directory name.')
    _parser.add_argument('--exclude', nargs='*', type=str, default=['figs', 'template_11701', 'slurm*'],
                         help='Directory names to exclude from sub-dirs (i.e. do not contain MESA runs).')
    _parser.add_argument('--out-file', '-o', type=str, default='../out_{}',
                         help='File format for MESA out files (from eg ./rn | tee out_file). Curly braces will be '
                              'expanded with with the directory name.')
    _parser.add_argument('--slurm-stats-dir', '-s', type=str, default='slurm_out',
                         help='Directory containing the slurm output (.err, .stats, .out files) for a job.')
    _parser.add_argument('--make-restart-file', '-r', type=str, default='',
                         help='Make a file which contains which subdir needs to restart from what photo.'
                              'Pass `pre-CHeX` to only restart runs which did not finish CHeX, pass `redoRC` to restart'
                              'before CHeB or `all` to do for all.')
    # _parser.add_argument('--good-termination-codes', '-t', nargs='*', type=list, default=['xa_central_lower_limit', 'stop_at_TP'], )
    _parser.add_argument('--no-slurm', action='store_const', const=True, default=False,
                         help='Run the check ignoring any slurm output.')
    _parser.add_argument('--list-all', action='store_const', const=True, default=False,
                         help='Print all termination reasons.')
    return _parser


def read_hist_first_last_row(fpath):
    with open(fpath, 'r') as f:
        for _ in range(6):
            columns = f.readline()
        first_row = f.readline()
        f.seek(0, 2)
        size = f.tell()
        f.seek(size - len(first_row)*2)
        last_row = f.readline()#.split('\n')[1]
    first_row = tuple(first_row.strip().split())
    last_row = tuple(last_row.strip().split())
    columns = columns.strip().split()
    formats = [np.array(ast.literal_eval(_)).dtype if _ != 'NaN' else np.float64 for _ in first_row]
    return np.rec.array([first_row, last_row], dtype=list(zip(columns, formats)))#dtype={'names': columns, 'formats': formats})


def get_termination_code(lines):
    # last 100 lines of MESA output
    lines = lines[-min(100, len(lines)):]
    term_code = 'Running'
    for line in lines[::-1]:
        if 'termination code' in line:
            term_code = line.split(':')[1].strip()
            break

    if term_code == 'Running':  # Didn't find normal termination code
        for line in lines:
            if 'STOP' in line:
                term_code = line.split('STOP')[1].strip()
                break
            if 'Backtrace' in line:
                term_code = 'Error'
            if ('adjust_mesh_failed' in line) or ('mesh_plan problem') in line:
                term_code = 'mesh_failed'
                break
            if 'stopping because of problems dt < min_timestep_limit' in line:
                term_code = 'min_timestep_limit'
                break
            if 'stop because hit EOS limits' in line:
                term_code = 'EOS_limits'
                break
            if 'stopping because of problems' in line:
                term_code = line.split('--')[-1].strip()
                break
            if 'terminated evolution: nonzero_ierr' in line:
                term_code = 'nonzero_ierr'

    return term_code


def get_timedelta(lines):
    first = lines[0].strip()
    last = lines[-1].strip()

    dts = []
    for timestamp in (first, last):
        try:
            _, day, month, hms, tz, year = timestamp.split()
        except ValueError:
            return dts[0] - dts[0]

        day = int(day)
        month = {'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 'May': 5, 'Jun': 6,
                 'Jul': 7, 'Aug': 8, 'Sep': 9, 'Oct': 10, 'Nov': 11, 'Dec': 12}[month]
        year = int(year)

        hour, minute, second = (int(_) for _ in hms.split(':'))
        dts.append(datetime(year, month, day, hour, minute, second))
    timedelta = (dts[1] - dts[0])

    return timedelta


def get_mesa_termcode(sub_dir, args):
    out_file = os.path.join(args.grid_dir, sub_dir, args.out_file).format(sub_dir)
    if os.path.exists(out_file):
        with open(out_file, 'r') as handle:
            out_file_lines = handle.readlines()

        # slurm_stats_file = os.path.join(args.grid_dir, sub_dir, args.slurm_stats_file).format(sub_dir)

        term_code = get_termination_code(out_file_lines)
        # walltime = get_timedelta(out_file_lines)
    else:
        term_code = 'NotRun'

    return sub_dir, term_code


def get_slurm_stats(stats_lines, args):
    info_dict = {}
    if args.no_slurm:
        info_dict['jobid'] = ''
        info_dict['jobname'] = 'noslurm'
        info_dict['job_task'] = ''
        info_dict['taskjob'] = ''
        info_dict['taskid'] = ''
        info_dict['walltime_requested'] = '24:0:0'
        info_dict['walltime_required'] = '24:0:0'
        info_dict['reason'] = 'None'
        return info_dict

    for line in stats_lines:
        if 'Identity jobid' in line:
            _, _, _, _jobid, _, _jobname = line.split()
            info_dict['jobid'] = _jobid
            info_dict['jobname'] = _jobname
        if 'Array Identity' in line:
            _, _, _, _job_task = line.split()
            _taskjob, _taskid = _job_task.split('_')
            info_dict['job_task'] = _job_task
            info_dict['taskjob'] = _taskjob
            info_dict['taskid'] = _taskid
        else:
            info_dict['job_task'] = ''
            info_dict['taskjob'] = ''
            info_dict['taskid'] = ''

        if 'Requested' in line:
            _, _, _hardware, _, _walltime_requested, _ = line.split()
            info_dict['hardware'] = _hardware
            info_dict['walltime_requested'] = _walltime_requested
        if 'Required' in line:
            _walltime_required = line.split()[-2]
            info_dict['walltime_required'] = _walltime_required
        if 'JobState' in line:
            _, _, _jobstate, _, _, _reason = line.split()
            info_dict['jobstate'] = _jobstate
            info_dict['reason'] = _reason

    return info_dict


def get_cheb_mask(data):
    """
    Defined as:
    have helium in core
    have convective core
    hydrogen depleted core


    :param data:
    :return:
    """

    mask = data.center_he4 > 1e-6
    mask = np.logical_and(mask, data.mass_conv_core > 0)
    mask = np.logical_and(mask, data.center_h1 < 1e-6)

    return mask


def count_in_dict(info_dict, key, filter_mainkeys=None):
    if filter_mainkeys is not None:
        filtered = dict((k, v[key]) for k, v in info_dict.items() if k in filter_mainkeys)
    else:
        filtered = dict((k, v[key]) for k, v in info_dict.items())
    cnt = Counter(filtered.values())
    dct = defaultdict(list)
    for main_key, value in filtered.items():
        dct[value].append(main_key)
    return cnt, dct


def invert_dict(dct, reference_key, filter_dct_keys=None, reverse_filter=False, sort=False):
    if filter_dct_keys is None:
        filter_dct_keys = []

    out_dict = defaultdict(list)
    for key, value in dct.items():
        if (key in filter_dct_keys) != reverse_filter:  # xor
            continue
        out_dict[value[reference_key]].append(key)

    if sort:
        for key in sorted(out_dict):
            out_dict[key] = sorted(out_dict[key])
    return out_dict




if __name__ == "__main__":
    run()

def run():
    parser = get_parser()
    args = parser.parse_args()

    args.grid_dir = os.path.abspath(args.grid_dir)

    # Expand items in args.exclude.
    _ = []
    for exclude in args.exclude:
        _ += glob.glob(os.path.join(args.grid_dir, exclude))
    args.exclude = _
    args.exclude = [os.path.split(_)[-1] for _ in args.exclude]

    # Get subdirs and remove items from args.exclude.
    if args.subdirs == '':
        subdirs = os.listdir(args.grid_dir)
        subdirs = sorted(subdirs)
        subdirs = [_ for _ in subdirs if os.path.isdir(os.path.join(args.grid_dir, _))]
        args.subdirs = subdirs

    _ = [glob.glob(os.path.join(args.grid_dir, _)) for _ in args.subdirs if _ not in args.exclude]
    args.subdirs = []
    for expanded in _:
        args.subdirs += expanded
    args.subdirs = [os.path.split(_)[-1] for _ in args.subdirs]

    # Get info from slurm stats files.
    jobid2task = {}
    task2jobid = {}
    run_info = {}
    run_data = {}
    slurm_stats_files = []
    if not args.no_slurm:
        if args.slurm_stats_dir == 'LOGS':

            for subdir in args.subdirs:
                for fname in os.listdir(os.path.join(args.grid_dir, subdir)):
                    if fname.endswith('.stats'):
                        slurm_stats_files.append((os.path.join(args.grid_dir, subdir, fname), subdir))
        else:
            for fname in os.listdir(os.path.join(args.grid_dir, args.slurm_stats_dir)):
                if fname.endswith('.stats'):
                    slurm_stats_files.append((os.path.join(args.grid_dir, args.slurm_stats_dir, fname), None))
            # slurm_stats_files = [os.path.join(args.grid_dir, args.slurm_stats_dir, _) )
    else:
        for subdir in args.subdirs:
            slurm_stats_files.append([None, subdir])
    for i, (slurm_stats_file, subdir) in enumerate(slurm_stats_files):
        if args.no_slurm:
            lines = None
        else:
            with open(slurm_stats_file, 'r') as handle:
                lines = handle.readlines()
        _ = get_slurm_stats(lines, args)
        jobid = _['jobid']
        taskjob = _['taskjob']
        taskid = _['taskid']
        if taskid == '':
            taskid = f'{i:>04}'

        if subdir is None:
            subdir = f'{taskid:>04}'

        if subdir in run_info:
            if int(run_info[subdir]['jobid']) < int(jobid):  # Take more recent run.
                run_info[subdir] = _
        else:
            run_info[subdir] = _

        jobid2task[jobid] = (taskjob, taskid)
        if taskid not in task2jobid:
            task2jobid[taskid] = jobid
        else:
            print(f'Key `{taskid}` already in `task2jobid` dictionary with value {task2jobid[taskid]} '
                  f'when inserting value {jobid}.')

    for subdir in args.subdirs:
        hist_path = os.path.join(args.grid_dir, subdir, f'LOGS/{args.history_file.format(subdir, subdir)}')
        mesa_termcode = get_mesa_termcode(subdir, args)[1]
        if os.path.exists(hist_path) and (mesa_termcode != 'NotRun'):
            data = read_hist_first_last_row(hist_path)
            run_info[subdir]['mesa_termcode'] = mesa_termcode
            run_info[subdir]['mesa_last_model'] = data.model_number[-1]
            # noinspection PyTypeChecker
            run_info[subdir]['CHeX'] = ['pre-CHeX', 'post-CHeX'][int(data.center_he4[-1] < 1e-6)]
            run_info[subdir]['CHX'] = ['pre-CHX', 'post-CHX'][int(data.center_h1[-1] < 1e-6)]
            _ = run_info[subdir]['walltime_required'].split(':')
            run_info[subdir]['short_runtime'] = ['', 'short_runtime'][
                int((3600 * int(_[0]) + 60 * int(_[1]) + int(_[0])) < 60)]

            run_data[subdir] = data
        else:
            run_info[subdir]['mesa_termcode'] = mesa_termcode
            run_info[subdir]['mesa_last_model'] = '-1'
            # noinspection PyTypeChecker
            run_info[subdir]['CHeX'] = mesa_termcode
            run_info[subdir]['CHX'] = mesa_termcode
            run_info[subdir]['short_runtime'] = mesa_termcode
            run_data[subdir] = mesa_termcode

    cnt_term_codes_mesa, dct_term_codes_mesa = count_in_dict(run_info, key='mesa_termcode')

    # Make output string.
    part_len = 44
    string_separator = '-' * part_len + '\n'
    status_str = '' + string_separator
    status_str += f'{"  termination_code":<32}{"count":>10}\n'
    status_str += string_separator
    for (mesa_tc, c_mesa_tc) in cnt_term_codes_mesa.items():
        part = f'{mesa_tc:<32}{c_mesa_tc:>12}\n'
        status_str += part
        if mesa_tc in ['None', 'min_timestep_limit']:
            cnt_term_codes_slurm, dct_term_codes_slurm = count_in_dict(run_info, key='reason',
                                                                       filter_mainkeys=dct_term_codes_mesa[mesa_tc])
            for i, (slurm_tc, c_slurm_tc) in enumerate(cnt_term_codes_slurm.items()):
                part = f'{slurm_tc:<29}{c_slurm_tc:>9}\n'
                start_part = ' ├ '
                if i == len(cnt_term_codes_slurm) - 1:
                    start_part = ' └ '
                part = start_part + part
                status_str += part

                # For each sub section count pre CHeX and post CHeX
                cnt_, dct_ = \
                    count_in_dict(run_info, key='CHeX', filter_mainkeys=dct_term_codes_slurm[slurm_tc])

                for j, key in enumerate(sorted(cnt_)):
                    part = f'{key:<26}{cnt_[key]:>6}\n'
                    start_part = '    ├ '
                    if j == len(cnt_) - 1:
                        start_part = '    └ '
                    part = start_part + part
                    status_str += part

                    # For each pre-CHeX section count pre CHX and post CHX
                    cnt_CHX, dct_CHX = \
                        count_in_dict(run_info, key='CHX', filter_mainkeys=dct_['pre-CHeX'])

                    for j, key in enumerate(sorted(cnt_CHX)):
                        part = f'{key:<23}{cnt_CHX[key]:>6}\n'
                        start_part = '       ├ '
                        if j == len(cnt_CHX) - 1:
                            start_part = '       └ '
                        part = start_part + part
                        status_str += part

    status_str += string_separator
    cnt, dct = count_in_dict(run_info, 'short_runtime')
    if not args.no_slurm:
        status_str += f'{"short_runtime":<32} {str(cnt["short_runtime"]):>12}\n'
        status_str += string_separator
    if args.list_all:
        for subdir in args.subdirs:
            status_str += f'{subdir:<16} {run_info[subdir]["CHeX"]:>10} {run_info[subdir]["mesa_last_model"]:>7} {run_info[subdir]["mesa_termcode"]:<12}\n'
        status_str += string_separator
    print(status_str)

    if args.make_restart_file:
        restart_reason = {}  # last photo, redoRC, etc
        for subdir in args.subdirs:
            info = run_info[subdir]
            data = run_data[subdir]
            reason = ''
            photo = ''
            mesa_termcode = info['mesa_termcode']

            # print(subdir, info['CHeX'], info['mesa_termcode'])
            if info['CHeX'] == 'post-CHeX':
                reason = 'redoRC'
                photo = 'preRC'
            elif mesa_termcode == 'Running':
                reason = 'Running'
                photo = 'last'
            elif info['CHeX'] == 'pre-CHeX':
                reason = 'None'
                photo = 'last'
            elif args.make_restart_file == 'redoRC':
                reason = 'redoRC'
                photo = 'preRC'
            elif mesa_termcode in (['logQ_min_limit', 'FileNotFound', 'NotRun']):
                reason = info['mesa_termcode']
                photo = 'full_restart'

            restart_reason[subdir] = (photo, reason)

        grid_restart_photos = []
        for subdir, (photo, reason) in restart_reason.items():
            photodir = os.path.join(args.grid_dir, subdir, 'photos')
            if os.path.exists(photodir):
                photos = sorted(os.listdir(photodir))
            else:
                photos = []
            photo_modelnum = np.array([int(_.replace('x', '')) for _ in photos])

            if args.make_restart_file == 'pre-CHeX':
                info = run_info[subdir]
                if info['CHeX'] == 'post-CHeX':
                    continue

            restart_photo = ''
            try:
                if photo == 'last':
                    restart_photo = photos[np.where((photo_modelnum[-1] - photo_modelnum) > 200)[0][-1]]
                elif photo == 'preRC':
                    histpath = os.path.join(args.grid_dir, subdir, f'LOGS/{args.history_file.format(subdir, subdir)}')
                    if os.path.exists(histpath):
                        hist = ld.History(histpath)
                        cheb_mask = uf.get_cheb_mask(hist)
                        first_model_rc = hist.data.model_number[cheb_mask][0]
                        restart_photo = photos[np.where((first_model_rc - photo_modelnum) > 200)[0][-1]]
                    else:
                        restart_photo = 'full_restart'
                elif photo == 'full_restart':
                    restart_photo = 'full_restart'
                elif photo == '':
                    restart_photo = 'full_restart'
            except IndexError as e:
                print(subdir, photo, reason, e)
                restart_photo = 'full_restart'
            if args.verbose:
                print(subdir, restart_photo)

            if restart_photo == '':
                print(f'Empty restart_photo for {subdir}!')
            grid_restart_photos.append((subdir, restart_photo))

        with open(os.path.join(args.grid_dir, 'grid_restart'), 'w') as f:
            f.writelines((' '.join(line) + '\n' for line in grid_restart_photos))
    return 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import dill
import joblib.externals.loky.process_executor
import numpy as np

from scipy import sparse
from scipy.spatial import Delaunay
from scipy import integrate as ig
from joblib import Parallel, delayed

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib import cm
from matplotlib.path import Path
from matplotlib.patches import PathPatch

# For debugging creation of connected zones
# Uncomment these imports, the make_path and plot_graph methods, and the associated lines in calc_zones
# import shapely as sh
# import networkx as nwx

from .. import functions as uf
from .. import load_data as ld
from . import utils as pu


class Kipp_data:
    def __init__(self, hist, profs, xaxis='model_number', yaxis='mass', caxis='eps_net', norm=None,
                 zone_filename='zones_wsssss.dat', verbose=False, save_zones=True, clobber_zones=False,
                 prof_prefix='profile', prof_suffix='.data', prof_resolution=500, parallel=True, ignore_monotonic=False):
        self.__version__ = '0.1.0'
        self.parallel = parallel
        self.verbose = verbose
        self.xaxis = xaxis
        self.yaxis = yaxis
        self.caxis = caxis
        self.xaxis_data = np.empty(len(hist)+1)
        self.xaxis_data[:-1] = hist.get(self.xaxis)
        self.xaxis_data[-1] = self.xaxis_data[-2] + (self.xaxis_data[-2] - self.xaxis_data[-3])
        self.has_mixtype = {}
        self.norm = norm
        self.color_info = None

        if os.name == 'nt':
            if parallel:
                print('Cannot use parallel on windows, settings to False.')
                self.parallel = False

        # Check if monotonic
        if not ignore_monotonic and not np.all(np.diff(np.sign(np.diff(self.xaxis_data))) == 0):
            raise ValueError(f'xaxis {xaxis} is not monotinically increasing or decreasing.')

        if zone_filename:
            self.zone_file = f'{hist.LOGS}/{zone_filename}'
        else:
            self.zone_file = ''
        self.load_zones = bool(self.zone_file)
        self.save_zones = save_zones and self.load_zones  # Also check if zone filename not empty
        self.prof_prefix = prof_prefix
        self.prof_suffix = prof_suffix
        self.prof_resolution = prof_resolution

        if yaxis == 'mass':
            self.mix_template_type = 'mix_type_{}'
            self.mix_template_top = 'mix_qtop_{}'
            self.burn_template_type = 'burn_type_{}'
            self.burn_template_top = 'burn_qtop_{}'
        elif yaxis == 'radius':
            self.mix_template_type = 'mix_relr_type_{}'
            self.mix_template_top = 'mix_relr_top_{}'
            self.burn_template_type = 'burn_relr_type_{}'
            self.burn_template_top = 'burn_relr_top_{}'
        else:
            self.mix_template_type = ''
            self.mix_template_top = ''
            self.burn_template_type = ''
            self.burn_template_top = ''

        if clobber_zones:
            self.load_zones = False

        if self.load_zones and os.path.exists(self.zone_file):
            if os.path.getmtime(self.zone_file) < os.path.getmtime(hist.path):
                self.load_zones = False
                if self.verbose:
                    print('zone file is older than loaded history file! Reloading.')
            else:
                self.load_zones = True
        else:
            self.load_zones = False

        burn_zones = None
        color_zones = None
        if self.load_zones:
            if self.verbose:
                print(f'Loading zonefile {self.zone_file}')
            try:
                loaded_version, loaded_yaxis, mixing_zones, loaded_caxis, burn_zones, yminmax = self.read_zones()
                self.ymin = yminmax[0]
                self.ymax = yminmax[1]
                if loaded_version != self.__version__:
                    if self.verbose:
                        print(f'Kipp_data version mismatch: {loaded_version} {self.__version__}')
                    load_success = False
                else:
                    load_success = True
                    if self.verbose:
                        print(f'Loading successfull. {len(mixing_zones)} mixing zones.')
            except Exception as exc:
                if self.verbose:
                    print(f'Loading zonefile failed. Recreating zones.')
                    print(exc)

                load_success = False

        if (not self.load_zones) or (not load_success):
            if self.caxis == 'eps_net':
                burn_zones = self.calc_zones(*self.get_hist_data(hist, 'burn'))
            mixing_zones = self.calc_zones(*self.get_hist_data(hist, 'mix'))
            self.load_zones = False

        if self.load_zones:
            # Check if loaded data is valid with inputs
            if loaded_yaxis != self.yaxis:
                if self.verbose:
                    print(f'yaxis from loaded zones and settings mismatch, recalculating yaxis and caxis: {loaded_yaxis} {yaxis}')
                mixing_zones = self.calc_zones(*self.get_hist_data(hist, 'mix'))
                if caxis == 'eps_net':
                    burn_zones = self.calc_color(hist, profs)
            elif loaded_caxis != self.caxis:
                if self.verbose:
                    print(f'caxis from loaded zones and settings mismatch, recalculating caxis: {loaded_caxis} {caxis}')
                if caxis == 'eps_net':
                    burn_zones = self.calc_color(hist, profs)
            else:  # Loaded correctly
                self.save_zones = False

        if self.parallel:
            joblib.externals.loky.get_reusable_executor().shutdown(wait=True)  # Kill workers

        self.mixing_zones = mixing_zones
        self.burn_zones = burn_zones

        if self.caxis not in ('', None, 'eps_net'):
            self.color_zones = self.calc_color(hist, profs, norm)
        else:
            self.color_zones = None
        if self.save_zones or clobber_zones:
            self.write_zones()

    def write_zones(self):
        if self.verbose:
            print(f'Writing zonefile {self.zone_file}')
        with open(self.zone_file, 'wb') as handle:
            dill.dump((self.__version__, self.yaxis, self.mixing_zones, self.caxis, self.burn_zones, (self.ymin, self.ymax)), handle)

    def read_zones(self):
        if self.verbose:
            print(f'Reading zonefile {self.zone_file}')
        with open(self.zone_file, 'rb') as handle:
            return dill.load(handle)

    def get_profile_xyz_data(self, hist, profs, norm):

        xyz_data = []
        for i, p in enumerate(profs):
            y = p.get(self.yaxis)
            c = p.get(self.caxis)

            #Normalize
            y_min = np.min(y)
            y_max = np.max(y)
            y_norm = (y - y_min)/(y_max - y_min)
            c_norm = norm(c)
            c_norm = c_norm.filled(np.nan)

            # Do a forwards and backwards pass to assist in triangulation
            idx = np.concatenate([pu.decimate_RDP(np.asarray([y_norm, c_norm]).T, epsilon=1/self.prof_resolution, return_index=True),
                                  pu.decimate_RDP(np.asarray([y_norm, c_norm]).T[::-1], epsilon=1 / self.prof_resolution, return_index=True)])
            idx = np.unique(idx)
            # If using norm's clip, remove first and last block of clipped values.
            if norm.clip:
                if (c_norm[idx[0]] <= 0 or (c_norm[idx[0]] >= 1)):
                    idx = idx[1:]
                if (c_norm[idx[-1]] <= 0 or (c_norm[idx[-1]] >= 1)):
                    idx = idx[:-1]

            # Insert 0 and 1 crossings and keep valid (0 < c < 1) points
            idx = np.concatenate([idx, np.where((c_norm[:-1] <= 0) & (c_norm[1:] > 0))[0]+1,  # Next good
                                       np.where((c_norm[:-1] > 0) & (c_norm[1:] <= 0))[0],  # Current good
                                       np.where((c_norm[:-1] < 1) & (c_norm[1:] >= 1))[0],  # Current good
                                       np.where((c_norm[:-1] >= 1) & (c_norm[1:] < 1))[0]+1])  # Next good
            idx = np.unique(idx)[::-1]

            xyz = np.zeros((3, idx.shape[0]))
            xyz[0] = p.get_hist_index(hist)
            xyz[1] = y[idx]
            xyz[2] = c[idx]
            xyz[2][(c_norm[idx] <= 0) | (c_norm[idx] >= 1)] = np.nan
            xyz_data.append(xyz)

        xyz_data = np.concatenate(xyz_data, axis=1)
        return xyz_data

    def get_hist_data(self, hist, kind):
        if kind == 'mix':
            template_top = self.mix_template_top
            template_type = self.mix_template_type
            bad_value = -1
        elif kind == 'burn':
            template_top = self.burn_template_top
            template_type = self.burn_template_type
            bad_value = -9999
        else:
            raise ValueError("kind must be either 'mix' or 'burn'.")

        constants = uf.get_constants(hist)
        center_column = ''
        if self.yaxis == 'mass':
            y_scale = hist.get('star_mass')

            if 'm_center' in hist.columns:
                center_column = 'm_center'
            elif 'm_center_gm' in hist.columns:
                center_column = 'm_center_gm'
                unit_scale = 1/constants.msun

        elif self.yaxis == 'radius':
            if 'r_center' in hist.columns:
                center_column = 'r_center'
                unit_scale = 1
            elif 'r_center_cm' in hist.columns:
                center_column = 'r_center_cm'
                unit_scale = 1/constants.rsun
            elif 'r_center_km' in hist.columns:
                center_column = 'r_center_km'
                unit_scale = 1e-3/constants.rsun

            if 'radius' in hist.columns:
                y_scale = hist.get('radius')
            elif 'log_R' in hist.columns:
                y_scale = 10**hist.get('log_R')
            else:
                print('Warning! Using photosphere radius as model radius.')
                y_scale = hist.get('photosphere_R')

        num_cols = len([c for c in hist.columns if c.startswith(template_top.format(''))])
        # for y_data, there is an implied column "0" where y=y_center (typically Y_center=0), so inlcude it explicitly
        y_data = np.zeros((num_cols+1, len(hist)))

        z_data = np.zeros((num_cols, len(hist)), dtype=int)
        for i in range(num_cols):
            y_data[i+1] = hist.get(template_top.format(i + 1)) * y_scale
            z_data[i] = hist.get(template_type.format(i + 1))
        if center_column != '':
            y_data[0] = unit_scale * hist.get(center_column)

        if kind == 'mix':
            z_data = uf.convert_mixing_type(z_data, hist.header['version_number'])

        self.ymax = y_scale.max()
        self.ymin = y_data[0].min()

        return y_data, z_data, bad_value

    def create_adjacency_matrix(self, y_data, z_data, bad_value):

        ny, nx = z_data.shape
        num_zones = ny * nx
        adjacency_matrix = sparse.dok_array((num_zones, num_zones), dtype=int)

        y_top = y_data[1:]
        y_bot = y_data[:-1]

        for i in range(ny):
            q_top = y_top[i]
            q_bot = y_bot[i]
            q_type = z_data[i]

            # zones must overlap in q and must be of same type and must not be bad_value
            connected_right = ~((q_top[np.newaxis, :-1] < y_bot[:, 1:]) | (q_bot[np.newaxis, :-1] > y_top[:, 1:])) & \
                               (q_type[np.newaxis, :-1] == z_data[:, 1:]) & (q_type[np.newaxis, :-1] != bad_value)
            i_y, i_x = np.where(connected_right)
            i_left = i_x + i*nx
            i_right = (i_x + 1) + i_y*nx
            for ii, ij in zip(i_left, i_right):
                adjacency_matrix[ii, ij] = 1  # Only need upper triangle as creating undirected graph

        return adjacency_matrix.tocsr()


    # def make_path(self, i, zone_ids, y_data, z_data):
    #     ny, nx = z_data.shape
    #     mask = zone_ids == i
    #     wmask = np.where(mask)[0]
    #     tops = y_data[1:].reshape((nx*ny))[mask]
    #     bots =  y_data[:-1].reshape((nx*ny))[mask]
    #     kind = z_data.reshape((nx*ny))[wmask[0]]
    #
    #     x = wmask % nx
    #     rects = np.zeros((len(x), 5, 2))
    #     rects[:, [0, 3, 4], 0] = x[:, np.newaxis]
    #     rects[:, [1, 2], 0] = x[:, np.newaxis] + 1
    #     rects[:, [0, 1, 4], 1] = tops[:, np.newaxis]
    #     rects[:, [2, 3], 1] = bots[:, np.newaxis]
    #
    #     rects = sh.polygons(rects)
    #     chunk_size = 32  # ~10% faster if doing chunks of rectangles
    #     poly = sh.union_all([sh.union_all(rects[i*chunk_size:(i+1)*chunk_size]) for i in range(len(rects)//chunk_size + 1)])
    #     # poly = sh.union_all(rects)
    #
    #     if isinstance(poly, sh.MultiLineString) or isinstance(poly, sh.MultiPolygon):  # Has holes
    #         vertices = []
    #         for sub_poly in poly.boundary.geoms:
    #             vertices.append(np.array(sub_poly.xy).T)
    #
    #         lengths = [len(_) for _ in vertices]
    #         num_vert = sum(lengths)
    #         vertices = np.row_stack(vertices)
    #         codes = np.zeros(num_vert)
    #         codes[:] = Path.LINETO
    #         codes[0] = Path.MOVETO
    #         for j in lengths[:-1]:
    #             codes[j + 1] = Path.MOVETO
    #         path = Path(vertices, codes)
    #     elif isinstance(poly, sh.Polygon):
    #         path = Path(np.array(poly.exterior.xy).T)
    #     else:
    #         print(i, type(poly))
    #         raise ValueError
    #     return kind, path


    def make_path2(self, zone_id, adjacency_matrix, zone_ids, y_data, z_data, return_indices=False):
        ny, nx = z_data.shape

        mask = zone_ids == zone_id
        i_adj = np.where(mask)[0]
        order_iadj = np.argsort(i_adj % nx, kind='stable')
        i_adj = i_adj[order_iadj]  # Sort by x then y

        kind = z_data.reshape((nx * ny))[i_adj[0]]
        tops = y_data[1:].reshape((nx * ny))[i_adj]
        bots = y_data[:-1].reshape((nx * ny))[i_adj]
        x = i_adj % nx

        num_nodes = len(i_adj)

        ll = 0  # lower left
        ul = 1  # upper left
        lr = 2  # lower right
        ur = 3  # upper right

        rects = np.zeros((num_nodes, 4, 2))
        rects[:, [ll, ul], 0] = x[:, np.newaxis]
        rects[:, [lr, ur], 0] = x[:, np.newaxis] + 1
        rects[:, [ll, lr], 1] = bots[:, np.newaxis]
        rects[:, [ur, ul], 1] = tops[:, np.newaxis]
        rects = rects.reshape((num_nodes * 4, 2))

        # More convenient indexing for vertices
        i4 = 4 * np.arange(num_nodes)

        un, i_x = np.unique(rects[:, 0], return_inverse=True)
        order = np.argsort(i_x, kind='stable')
        change_x = np.where(np.diff(i_x[order], prepend=-1, append=un[-1]+1))[0]

        # need int32 as otherwise sparse.csgraph.shortestpath breaks
        indices = np.zeros((2, num_nodes*4), dtype=np.int32)
        # Horizontal indices
        indices[:, :num_nodes] = i4 + ll, i4 + lr
        indices[:, num_nodes:num_nodes * 2] = i4 + ur, i4 + ul

        # Vertical indices
        no_holes = True
        max_num = 0
        i_ind_start = 2 * num_nodes
        for i in range(len(change_x)-1):
            i_start = change_x[i]
            i_end = change_x[i + 1]
            num = i_end - i_start
            i_ind_end = i_ind_start + num//2
            if no_holes:
                max_num = max(max_num, num)
            if no_holes and (num < max_num):
                no_holes = False

            o = order[i_start:i_end]
            v = o[np.argsort(rects[o, 1])]  # vertical order
            indices[:, i_ind_start:i_ind_end] = v[::2], v[1::2]
            i_ind_start = i_ind_end

        # Get correct directions for building Path
        # even nodes go to higher node number, odd nodes go to lower node number
        # if mixed, ul -> ll and lr -> ur if the corners belong to the same zone
        # otherwise it is the opposite
        is_odd = np.logical_and(*(indices%2).astype(bool))  # both odd
        is_even = np.logical_and(*np.logical_not((indices % 2).astype(bool)))  # both even
        is_both = ~np.logical_xor(is_even, is_odd)  # one odd one even
        is_left = is_both & (indices[0] % 4 <= 1)
        is_right = is_both & (indices[0] % 4 >= 2)
        is_same = (np.diff(indices//4, axis=0) == 0)[0]

        odd_like = is_odd | (is_right & ~is_same) | (is_left & is_same)
        even_like = is_even | (is_left & ~is_same) | (is_right & is_same)
        indices[:, odd_like] = np.sort(indices[:, odd_like], axis=0)[::-1]
        indices[:, even_like] = np.sort(indices[:, even_like], axis=0)

        if return_indices:
            return rects, indices, no_holes
        return kind, self.path_from_adjacency_matrix(indices, rects, no_holes)


    def path_from_adjacency_matrix(self, indices, nodes, no_holes):
        adjacency_matrix = sparse.coo_array((np.ones(len(nodes), dtype=np.int32), indices),
                                   shape=(len(nodes), len(nodes))).tocsr()
        n_comp, labels = sparse.csgraph.connected_components(adjacency_matrix, return_labels=True)
        path_indices = [np.where(labels==i_path)[0] for i_path in range(n_comp)]
        starts = [pathi[0] for pathi in path_indices]
        dist_mat = sparse.csgraph.shortest_path(adjacency_matrix, unweighted=True, indices=starts)

        offsets = np.cumsum([0] + [len(pathi) for pathi in path_indices])

        order = np.zeros(len(nodes) + n_comp, dtype=int)
        codes = np.full(len(order), Path.LINETO)
        for i in range(n_comp):
            order[offsets[i]+i:offsets[i+1]+i] = np.argsort(dist_mat[i], kind='stable')[:offsets[i+1]-offsets[i]]
            order[offsets[i+1]+i] = order[offsets[i]+i]
        vertices = nodes[order]
        codes[offsets[:-1] + np.arange(n_comp, dtype=int)] = Path.MOVETO

        # Remove double points
        mask = np.sum(np.abs(np.diff(vertices, axis=0, prepend=[[1e99, 1e99],])), axis=1) != 0
        vertices = vertices[mask]
        codes = codes[mask]
        return Path(vertices, codes)

    def calc_zones(self, y_data, z_data, bad_value):
        self.bad_value = bad_value
        ny, nx = z_data.shape
        adjacency_matrix = self.create_adjacency_matrix(y_data, z_data, bad_value)

        _, zone_ids = sparse.csgraph.connected_components(adjacency_matrix, directed=False)
        zone_ids[z_data.reshape(nx*ny) == bad_value] = -1
        zone_ids = np.digitize(zone_ids, bins=np.unique(zone_ids)) - 2

        num_zones = np.max(zone_ids) + 1

        path_function = self.make_path2

        if self.parallel:
            try:
                zones = Parallel(n_jobs=-1)(
                    delayed(self.make_path2)(i, adjacency_matrix, zone_ids, y_data, z_data) for i in range(num_zones))
                # zones = Parallel(n_jobs=1)(
                #     delayed(self.make_path)(i, zone_ids, y_data, z_data) for i in range(num_zones))
            except joblib.externals.loky.process_executor.TerminatedWorkerError as exc:
                print('Parallel failed, falling back to single threaded.')
                zones = [self.make_path2(i, adjacency_matrix, zone_ids, y_data, z_data) for i in range(num_zones)]
        else:
            zones = [self.make_path2(i, adjacency_matrix, zone_ids, y_data, z_data) for i in range(num_zones)]
            # zones = [self.make_path(i, zone_ids, y_data, z_data) for i in range(num_zones)]
        return zones


    def add_mixing(self, ax, xlims, ylims, mixing_min_height, kwargs_mixing=None):
        if self.verbose:
            print('Adding mixing to axis.')
        if kwargs_mixing is None:
            kwargs_mixing = pu.get_default_mixing_kwargs()

        self.has_mixtype = {mix_type:False for mix_type in kwargs_mixing.keys()}

        min_ix = 1e99
        max_ix = -1e99
        get_xlim = True
        if xlims is not None:
            min_ix = np.argwhere(xlims[0] < self.xaxis_data)[0][0]
            min_ix = max(0, min_ix - 1)
            max_ix = np.argwhere(xlims[1] > self.xaxis_data)[0][-1]
            max_ix = min(len(self.xaxis_data) - 1, max_ix + 1)
            get_xlim = False

        if get_xlim:
            if self.color_zones is not None:
                if isinstance(self.color_zones, list):
                    for _, path in self.color_zones:
                        min_ix = min(min_ix, min(path.vertices[:, 0]))
                        max_ix = max(max_ix, max(path.vertices[:, 0]))
                elif isinstance(self.color_zones, np.ndarray):
                    x = self.color_zones[0]  # Only need 1 column
                    min_ix = min(min_ix, min(x))
                    max_ix = max(max_ix, max(x))
            else:
                for _, path in self.mixing_zones:
                    min_ix = min(min_ix, min(path.vertices[:, 0]))
                    max_ix = max(max_ix, max(path.vertices[:, 0]))

        patches_dict = {key:[] for key in kwargs_mixing.keys()}
        patches_dict['order'] = np.zeros((len(kwargs_mixing), 2), dtype=int)
        patches_dict['order'][:,1] = -1  # If still -1 then none were added
        mix_type_index = dict(zip(kwargs_mixing.keys(), np.arange(len(kwargs_mixing))))
        for i, (mix_type, path) in enumerate(self.mixing_zones):
            if mix_type in kwargs_mixing.keys():
                mix_info = kwargs_mixing[mix_type]
                color = mix_info['color']
                hatch = mix_info['hatch']
                line = mix_info['line']
                show = mix_info['show']
                if not show:
                    continue
            else:
                continue

            min_ix = int(min_ix)
            max_ix = int(max_ix)
            x_extent = self.xaxis_data[[min_ix, max_ix]]

            # Skip zones outside xlims
            if path.vertices[:,0].min() > max_ix:
                continue
            if path.vertices[:,0].max() < min_ix:
                continue
            if mixing_min_height > 0:
                if path.vertices[:,1].max() - path.vertices[:,1].min() < mixing_min_height:
                    continue

            # Convert hist index coords to x-data coords
            new_vert = path.vertices.copy()
            new_vert[:,0] = self.xaxis_data[path.vertices[:,0].astype(int)]
            path = Path(new_vert, path.codes)

            # ax.add_patch(PathPatch(path, fill=False, hatch=hatch, edgecolor=color, linewidth=line))
            patches_dict[mix_type].append(PathPatch(path, fill=False, hatch=hatch, edgecolor=color, linewidth=line))
            patches_dict['order'][mix_type_index[mix_type]] = mix_type, i
            self.has_mixtype[mix_type] = True
        patches_dict['order'] = patches_dict['order'][patches_dict['order'][:,1] != -1]  # Remove unused mix_types
        ordered_mixtypes = patches_dict['order'][:,0][np.argsort(patches_dict['order'][:,1])]

        for mix_type in ordered_mixtypes:
            patches = patches_dict[mix_type]
            mix_info = kwargs_mixing[mix_type]
            color = mix_info['color']
            hatch = mix_info['hatch']
            line = mix_info['line']
            show = mix_info['show']
            if not show:
                continue
            ax.add_collection(mpl.collections.PatchCollection(patches, match_original=True, hatch=hatch, edgecolor=color, linewidth=line))
        return x_extent

    def add_color(self, ax, xlims, ylims, clims, norm=None, cmap=None, kwargs_profile_color=None):
        if self.verbose:
            print('Adding color to axis.')
        min_ix = 1e99
        max_ix = -1e99
        get_xlim = True
        if xlims is not None:
            min_ix = np.argwhere(xlims[0] < self.xaxis_data)[0][0]
            min_ix = max(0, min_ix - 1)
            max_ix = np.argwhere(xlims[1] > self.xaxis_data)[0][-1]
            max_ix = min(len(self.xaxis_data)-1, max_ix+1)
            get_xlim = False

        if (self.burn_zones is not None) and self.caxis == 'eps_net':  # Colors from burn_type_* from history.
            if clims is None:
                vmin = min([_[0] for _ in self.burn_zones])
                vmax = max([_[0] for _ in self.burn_zones])
            else:
                vmin, vmax = clims

            if norm is None:
                norm = pu.MidpointBoundaryNorm(np.linspace(vmin, vmax, int(vmax - vmin + 1)), 256, 0)
            if cmap is None:
                cmap = pu.cm.RdBu

            self.color_info = (norm, cmap)

            ax.set_facecolor(cmap(0.5))
            if get_xlim:
                for burn_type, path in self.burn_zones:
                    min_ix = min(min_ix, min(path.vertices[:, 0]))
                    max_ix = max(max_ix, max(path.vertices[:, 0]))
            min_ix = int(min_ix)
            max_ix = int(max_ix)
            x_extent = self.xaxis_data[[min_ix, max_ix]]

            patches = []
            for burn_type, path in self.burn_zones:
                # Keep no/very low burning as middle color and skip drawing as it is already the background color
                if burn_type == 0 and cmap is pu.cm.RdBu:
                    continue
                # Skip zones outside xlims
                if path.vertices[:,0].min() > max_ix:
                    continue
                if path.vertices[:,0].max() < min_ix:
                    continue

                # Convert hist index coords to x-data coords
                new_vert = path.vertices.copy().astype(float)
                new_vert[:, 0] = self.xaxis_data[path.vertices[:, 0].astype(int)]

                path = Path(new_vert, path.codes)
                patches.append(
                    PathPatch(path, fill=True, edgecolor=None, color=cmap(norm(burn_type)),
                              zorder=burn_type - len(self.burn_zones)))
            ax.add_collection(mpl.collections.PatchCollection(patches, match_original=True))

        else:
            x, y, c = self.color_zones
            x = self.xaxis_data[x.astype(int)]

            if clims is None:
                vmin = np.nanmin(c)
                vmax = np.nanmax(c)
            else:
                vmin, vmax = clims

            if kwargs_profile_color is None:
                kwargs_profile_color = {'shading': 'gouraud'}
            else:
                shading = {'shading': 'gouraud'}
                shading.update(kwargs_profile_color)
                kwargs_profile_color = shading

            if norm is not None:
                if 'norm' in kwargs_profile_color.keys():
                    print(f'Using norm from add_color argument.')
                kwargs_profile_color['norm'] = norm
            else:
                if 'norm' not in kwargs_profile_color.keys() or kwargs_profile_color['norm'] is None:
                    kwargs_profile_color['norm'] = mpl.colors.Normalize(vmin, vmax)
                    norm = kwargs_profile_color['norm']
            if cmap is not None:
                if 'cmap' in kwargs_profile_color.keys():
                    print(f'Using cmap from add_color argument.')
            kwargs_profile_color['cmap'] = cmap

            triangulation_pts = self.color_zones.copy().T
            all_simplices = self._triangulate(triangulation_pts[:, :2])

            # if any vertex contains a nan, remove triangles containing it
            finite = np.all(np.isfinite(triangulation_pts[all_simplices]), axis=(1, 2))
            all_simplices = all_simplices[finite]

            # Retriangulate connected areas
            adjacency_matrix = sparse.dok_array((self.color_zones.shape[1], self.color_zones.shape[1]), dtype=int)
            adjacency_matrix[all_simplices[:, 0], all_simplices[:, 1]] = 1
            adjacency_matrix[all_simplices[:, 1], all_simplices[:, 2]] = 1
            adjacency_matrix[all_simplices[:, 2], all_simplices[:, 0]] = 1

            adjacency_matrix = adjacency_matrix.tocsr()
            n_components, labels = sparse.csgraph.connected_components(adjacency_matrix, directed=False)
            cts, _ = np.histogram(labels, np.arange(n_components))

            all_triangles = []
            main_index = np.arange(len(triangulation_pts), dtype=int)
            for good_label in np.where(cts > 2)[0]:
                mask = labels == good_label
                # triangles = self._triangulate(triangulation_pts[mask, :2], cumulative_integral=False)
                triangles = self._triangulate(triangulation_pts[mask], cumulative_integral=True)

                ax.tripcolor(x.flat[mask], y.flat[mask], c.flat[mask], triangles=triangles, **kwargs_profile_color)
                # ax.triplot(x.flat[mask], y.flat[mask], triangles=triangles, marker='.')
                triangles = main_index[mask][triangles]  # Unmask indeces
                all_triangles.append(triangles)
            x_extent = np.array([x[0], x[-1]])

            self.color_info = (kwargs_profile_color['norm'], kwargs_profile_color['cmap'])
            self.triangulation_pts = triangulation_pts
            self.simplices = all_triangles

        return x_extent

    def _triangulate(self, pts, cumulative_integral=False):
        """
        Normalize and triangulate `pts`.

        Args:
            pts:

        Returns:
            Triangulation simplices.
        """
        if not np.all(np.isfinite(pts)):
            raise ValueError(f'Non-finite value in triangulation points.')

        unique_x, cts = np.unique(pts[:,0], return_counts=True)
        i_start_end = np.cumsum([0, *cts])
        i_start_end[-1] = -1

        pts[:,0] = np.digitize(pts[:,0], unique_x, right=True)

        pts = (pts - pts.min(axis=0)) / (pts.max(axis=0) - pts.min(axis=0))

        if cumulative_integral:
            for i in range(len(cts)):
                i_start = i_start_end[i]
                i_end = i_start_end[i + 1]
                pts[i_start:i_end, 2] = ig.cumulative_trapezoid(pts[i_start:i_end, 2], pts[i_start:i_end, 1], initial=0)
                if i_end - i_start > 2:
                    pts[i_start:i_end, 2] -= pts[i_start+1, 2]
                pts[i_start:i_end, 2] /= pts[i_end-1, 2]
            pts = pts[:, [0,2]]  # Remove 3rd dimension for triangulation.

        all_simplices = []
        for i in range(len(cts) - 1):  # Generate triangulation
            i_start = i_start_end[i]
            i_end = i_start_end[i + 2]
            delan = Delaunay(pts[i_start:i_end])
            all_simplices.append(delan.simplices + np.sum(cts[:i]))
        all_simplices = np.concatenate(all_simplices)
        return all_simplices

    def make_kipp(self, ax=None, xlims=None, ylims=None, clims=None, norm=None, cmap=None, mixing_min_height=0, kwargs_mixing=None,
                  kwargs_profile_color=None):
        if self.verbose:
            print(f'Making figure.')

        f, ax = pu.get_figure(ax)

        mixing_min_height *= (self.ymax - self.ymin)
        if (self.color_zones is not None or self.burn_zones is not None) and self.caxis:
            c_extent = self.add_color(ax, xlims, ylims, clims, norm, cmap, kwargs_profile_color)
        else:
            c_extent = [1e99, -1e99]
        m_extent = self.add_mixing(ax, xlims, ylims, mixing_min_height, kwargs_mixing)

        xextent = np.zeros(2)
        xextent[0] = min(c_extent[0], m_extent[0])
        xextent[1] = max(c_extent[1], m_extent[1])
        ax.set_xlim(xextent)

        if ylims is None:
            ax.set_ylim(self.ymin, self.ymax)

        if self.verbose:
            print('Done.')
        return f, ax

    # def plot_graph(self, adjacency_matrix, z_data, bad_value):
    #     ny, nx = z_data.shape
    #     _, zone_ids = sparse.csgraph.connected_components(adjacency_matrix, directed=False)
    #     mask_bad = z_data.reshape(nx*ny) == bad_value
    #     if np.any(mask_bad):
    #         zone_ids[mask_bad] = -1
    #         zone_ids = np.digitize(zone_ids, bins=np.unique(zone_ids)) - 2
    #     else:
    #         zone_ids = np.digitize(zone_ids, bins=np.unique(zone_ids)) - 2
    #     colors = np.digitize(z_data.reshape(nx*ny), bins=np.unique(z_data))
    #
    #     f, ax = plt.subplots()
    #     ax.imshow(zone_ids.reshape(z_data.shape), aspect='auto', origin='lower')
    #     G = nwx.from_numpy_array(adjacency_matrix)
    #     nwx.draw_networkx(G, pos={node: (node % nx, node // nx) for node in G.nodes},
    #                       node_color=colors,
    #                       with_labels=False, ax=ax,
    #                       node_size=32, hide_ticks=False)

    def calc_color(self, hist, profs, norm=None):
        if self.caxis == 'eps_net' and 'burn_qtop_1' in hist.columns:
            return self.calc_zones(*self.get_hist_data(hist, 'burn'))

        if ((profs is None) or len(profs) == 0):
            profs = ld.load_profs(hist, prefix=self.prof_prefix, suffix=self.prof_suffix)
            if len(profs) == 0:
                return None
            # Check if profiles have required columns
            # required_columns = []
            # if self.yaxis == 'mass':
            #     required_columns.append('mass')
            # elif self.yaxis == 'radius':
            #     required_columns.append('radius')
            # required_columns.append(self.caxis)
            required_columns = [self.yaxis, self.caxis]
            missing_cols = []
            for p in profs:
                for c in required_columns:
                    if not c in p.columns:
                        missing_cols.append((c, p.fname))

            if len(missing_cols) > 0:
                missing_cols_str = ''
                for c, pname in missing_cols:
                    missing_cols_str += f'{pname} {c}\n'
                raise ValueError(f'Missing required columns in profile files:\n'
                                 f'{missing_cols_str}')

        vmin = 1e99
        vmax = -1e99
        for p in profs:
            c = p.get(self.caxis)
            if isinstance(norm, mpl.colors.LogNorm):
                vmin = min(vmin, np.min(c[c>0]))  # Smallest positive
            else:
                vmin = min(vmin, np.min(c))
            vmax = max(vmax, np.max(c))

        if norm is not None:
            if norm.vmin is not None:
                vmin = norm.vmin
            if norm.vmax is not None:
                vmax = norm.vmax
            norm = norm.__class__(vmin=vmin, vmax=vmax, clip=norm.clip)
        else:
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

        return self.get_profile_xyz_data(hist, profs, norm)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""MESA grid overseer."""
import functools
import itertools
import multiprocessing as mp
import os
import signal
import subprocess
import sys
import time
from argparse import ArgumentParser
import numpy as np

def get_parser():
    """"""
    # TODO: Add option to compile star once and then copy the executable?
    parser = ArgumentParser(description="""
    This program will run a MESA grid created with ``wsssss.inlists.create_grid`` module.
    """)
    try:
        env_OMP_NUM_THREADS = os.environ['OMP_NUM_THREADS']
    except KeyError:
        print('Environment variable ``OMP_NUM_THREADS`` not set, using ``nproc``.')
        env_OMP_NUM_THREADS = mp.cpu_count()

    parser.add_argument('grid_dir', type=str, default='.', nargs='?',
                        help='The directory which contains each run for MESA to run in a separate sub-directory.')
    parser.add_argument('--sub-dirs', '-d', nargs='*', type=str, default='',
                        help='The names of the sub directories if they are not named 0000 0001 etc.')
    parser.add_argument('--base-work-dir', '-b', type=str, default='$MESA_DIR/star/work',
                        help='Directory to use as the base work directory. A copy will be made and ``star`` executed '
                             'in this directory.')
    parser.add_argument('--restart', '-re', default=False, const=True, nargs='?',
                        help='Restart from the last photo in each run\'s photos directory. If no photo is found, '
                             'start a new run. If a check-grid grid_restart file is passed, use that instead.')
    parser.add_argument('--source', type=str, default='',
                        help='Source this file before running ``star``.')
    parser.add_argument('--cmd-pre', type=str, default='',
                        help='Command to run before starting the grid. Can be a file.')
    parser.add_argument('--cmd-pre-each', type=str, default='',
                        help='Command to run directly before starting ``star`` in each directory. Can be a file.')
    parser.add_argument('--cmd-main', type=str, default='./rn',
                        help='Command to run MESA. Defaults to ``./rn``.')
    parser.add_argument('--cmd-post-each', type=str, default='',
                        help='Command to run directly after ``star`` has finished in each directory. Can be a file.')
    parser.add_argument('--cmd-post', type=str, default='',
                        help='Command to run after the grid is finished. Can be a file.')
    parser.add_argument('--verbose', '-v', action='store_const', const=True, default=False,
                        help='Print info about what the script is doing.')
    parser.add_argument('--num-mesa', '-n', type=int, default=1,
                        help='Number of instances of MESA to run at once.')
    parser.add_argument('--OMP_NUM_THREADS', '-j', type=int, default=env_OMP_NUM_THREADS,
                        help='Value of the OMP_NUM_THREADS environment variable to set before running ``star``. '
                             '``num-mesa`` times ``OMP_NUM_THREADS`` cannot be larger than ``nproc``. '
                             'Defaults to `multiprocessing.cpu_count()` if `OMP_NUM_THREADS` environment variable '
                             'is not set.')
    parser.add_argument('--debug', action='store_const', const=True, default=False,
                        help='Print info about what the script is doing.')
    parser.add_argument('--skip-if-file-exists', type=str, default='',
                        help='Skip running MESA in this work_dir if this file exists in work_dir. '
                             'Can use MESAGO_LOG_FILE to look for log file generated by this script.'
                             'WORK_DIR and RUN_NAME will be replaced with the ``run_name``, so '
                             'WORK_DIR/out_RUN_NAME becomes $GRID_DIR/0000/out_0000.')
    parser.add_argument('--log-path', type=str, default='out_RUN_NAME',
                        help='Log file path, WORK_DIR, and RUN_NAME will be replaced by their '
                             'corresponding values.')
    parser.add_argument('--task-share', action='store_const', const=True, default=False,
                        help='Set the slurm task-id and only do runs assigned to that task-id.')
    return parser


def check_cores(args):
    """Check and process args."""
    nproc = os.cpu_count()
    req_cores = args.num_mesa * args.OMP_NUM_THREADS

    if req_cores > nproc:
        raise ValueError(f'Number or required cores ({req_cores}) larger than available cores ({nproc}).')
    return req_cores, nproc


def process_args(args):
    """"""

    def expand_path(path):
        return os.path.abspath(os.path.expandvars(path))

    args.base_work_dir = expand_path(args.base_work_dir)
    args.grid_dir = expand_path(args.grid_dir)

    if args.source != '':
        args.source = expand_path(args.source)

    if args.log_path != '':
        if not (('WORK_DIR' in args.log_path) or ('RUN_NAME' in args.log_path)):  #
            raise ValueError('log-path must contain one or both of WORK_DIR or RUN_NAME in its path.')

    if args.skip_if_file_exists != '':
        if not (('WORK_DIR' in args.log_path) or ('RUN_NAME' in args.log_path)):  #
            raise ValueError('skip-if-file-exists must contain one or both of WORK_DIR or RUN_NAME in its path.')

    args.restart_settings = None
    if isinstance(args.restart, str):
        with open(args.restart, 'r') as handle:
            lines = handle.readlines()
        restart_settings = {}
        for line in lines:
            dirname, photo = line.split()
            restart_settings[dirname] = photo
        args.restart_settings = restart_settings
        args.restart = True
    return args


def start_mesa(args, run_name, logger):
    pid = os.getpid()
    logger.info(f'{pid}: Starting {run_name}')

    os.chdir(os.path.join(args.grid_dir, run_name))
    if args.verbose:
        logger.info(f'{pid}: Current directory: {os.getcwd()}')

    if args.log_path == '':
        log_file = os.path.join(args.grid_dir, 'out_' + run_name)
    else:
        log_file = os.path.join(args.grid_dir, args.log_path). \
            replace('WORK_DIR', run_name).replace('RUN_NAME', run_name)
    os.makedirs(os.path.dirname(log_file), exist_ok=True)

    if args.skip_if_file_exists:
        if args.skip_if_file_exists == 'MESAGO_LOG_FILE':
            skip_check_fpath = log_file
        else:
            skip_check_fpath = os.path.join(args.grid_dir, args.skip_if_file_exists).replace('WORK_DIR', run_name). \
                replace('RUN_NAME', run_name)

        if os.path.exists(skip_check_fpath):
            logger.info(f'{pid}: Skipping {run_name}, skip file found {skip_check_fpath}.')
            return run_name, 'Skip file found.'

    with open(log_file,
              'w') as handle:  # Create file so other processes can find it before someting actually writes to it
        handle.write('')
        handle.flush()
        os.fsync(handle)

    pre_cmd_str = f'export OMP_NUM_THREADS={args.OMP_NUM_THREADS}; '
    if args.source != '':
        pre_cmd_str = f'. {args.source}; ' + pre_cmd_str

    if args.cmd_pre_each != '':
        run_cmd(args.cmd_pre_each, shell=True)

    if not os.path.exists('star'):
        run_cmd('./clean && ./mk', shell=True)

    if args.debug:
        out = run_cmd(f'{os.path.dirname(os.path.abspath(__file__))}/testing_star.py {run_name}')
        if args.cmd_post_each != '':
            run_cmd(args.cmd_post_each)
        return run_name, out

    if os.path.exists('photos'):
        photos = os.listdir('photos')
    else:
        photos = []
    if args.restart and len(photos) > 0:
        if args.restart_settings is None:
            photos_int = [photo.replace('x', '0') for photo in photos]
            photos_order = np.argsort(np.array([int(photo) for photo in photos_int if photo.isdigit()]))
            photos = np.array(photos)[photos_order].tolist()
            photo = photos[-1]
        else:
            photo = args.restart_settings[run_name]
        if photo == 'full_restart':
            cmd = pre_cmd_str + args.cmd_main + '  2>&1'
            if args.verbose:
                logger.info(f'{pid}: {run_name} {cmd}')
            out = run_cmd(cmd, split=False, shell=True, to_file=log_file)
        else:
            cmd = pre_cmd_str + f'./re {photo} >> {log_file} 2>&1'
        if args.verbose:
            logger.info(f'{pid}: {run_name} {cmd}')
        out = run_cmd(cmd, split=False, shell=True, to_file=log_file)
    else:
        cmd = pre_cmd_str + args.cmd_main + '  2>&1'
        if args.verbose:
            logger.info(f'{pid}: {run_name} {cmd}')
        out = run_cmd(cmd, split=False, shell=True, to_file=log_file)

    if args.cmd_post_each != '':
        run_cmd(args.cmd_post_each, shell=True)

    logger.info(f'{pid}: Finished {run_name}')

    return run_name, out

def queue_start_mesa(queue, logger):
    while True:
        item = queue.get(block=True)
        if item is None:
            break

        start_mesa(*item, logger=logger)


def run_cmd(cmd, capture_output=False, split=False, to_file='', file_mode='w', **kwargs):
    pid = os.getpid()
    if file_mode not in ['a', '-a', 'w']:
        raise ValueError('`file_mode` must be one of `a`, `-a`, or `w`.')

    if capture_output and to_file != '':
        logger.warning(f'{pid}: Cannot simultaneously capture output and write to file if `split=True`, setting to False.')
        split = False

    if split:
        cmd = cmd.split()

    if to_file != '' and not capture_output:
        with open(to_file, file_mode) as f:
            return subprocess.run(cmd, stdout=f, stdin=f, **kwargs)
    if capture_output:
        if to_file != '':
            if file_mode == 'a':
                file_mode = '-a'
            elif file_mode == 'w':
                file_mode = ''
            return subprocess.run(cmd + f'| tee {file_mode} {to_file}', stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  shell=True, **kwargs)
        return subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
    return subprocess.run(cmd, **kwargs)


def get_subdirs(args):
    if args.sub_dirs != '':
        sub_dirs = args.sub_dirs
    else:
        sub_dirs = []
        for item in os.listdir(args.grid_dir):
            item_path = os.path.join(args.grid_dir, item)
            if os.path.isdir(item_path) and item.isdigit():
                sub_dirs.append(item)
        sub_dirs.sort()
    return sub_dirs


def main(args, logger):
    # TODO: include walltime?

    if args.verbose:
        print('Parsed inputs:')
        for key, value in args.__dict__.items():
            print(f'{key:<16}= {value}')
        print('')

    check_cores(args)

    if args.cmd_pre != '':
        run_cmd(args.cmd_pre, shell=True)

    sub_dirs = get_subdirs(args)

    arguments = list(zip(itertools.repeat(args), sub_dirs))
    if args.task_share:
        try:
            n_tasks = int(os.environ['SLURM_ARRAY_TASK_COUNT'])
            task_id = int(os.environ['SLURM_ARRAY_TASK_ID'])
            task_id_min = int(os.environ['SLURM_ARRAY_TASK_MIN'])
            task_id = task_id - task_id_min
        except KeyError:  # Not running with a task array
            n_tasks = 1
            task_id = 0
        print(f'n_tasks={n_tasks}')
        print(f'task_id={task_id}')
    else:
        n_tasks = 1
        task_id = 0

    if args.verbose:
        print('Sub-directories:')
        for subdir in sub_dirs[task_id::n_tasks]:
            print(subdir)
        print('')

    _queue_start_mesa = functools.partial(queue_start_mesa, logger=logger)

    queue = mp.Queue()
    pool = mp.Pool(args.num_mesa, _queue_start_mesa, (queue,))
    for item in arguments[task_id::n_tasks]:
        queue.put(item)
    for i in range(args.num_mesa):
        queue.put(None)
    queue.close()
    queue.join_thread()
    pool.close()
    pool.join()

    if args.cmd_post != '':
        run_cmd(args.cmd_post, shell=True)


def run():
    try:
        import setproctitle

        setproctitle.setproctitle('MESAgo')
    except ModuleNotFoundError:
        print('Module setproctitle not found, will not set process name.')
        pass

    signal.signal(signal.SIGTERM, signal.default_int_handler)
    signal.signal(signal.SIGINT, signal.default_int_handler)

    t_start = time.time()

    parser = get_parser()
    args = parser.parse_args()

    args = process_args(args)

    logger = mp.get_logger()

    if args.debug:
        args.verbose = True
        logger.setLevel('DEBUG')
    else:
        logger.setLevel('INFO')

    main(args, logger)
    return 0


if __name__ == "__main__":
    run()
